#include <iostream>
#include <fstream>
#include <vector>
#include <ctime>
#include <sstream>

#include "rng.h"
#include "tree.h"
#include "info.h"
#include "funs.h"
#include "calibd.h"

using std::cout;
using std::endl;

int main(int argc, char** argv)
{

	//Load the fitted trees and cutpoints
	cout << "Loading trees" << endl;
	size_t m,p,ndraws;
	xinfo xi;
	std::ifstream treef("trees.txt");
	treef >> xi; //load the cutpoints
	treef >> m;  //number of trees
	treef >> p;  //dimension of x's
	treef >> ndraws; //number of draws from the posterior that were saved.

	cout << "Loaded " << ndraws << " draws saved from tree posterior" << endl;
	std::vector<std::vector<tree> > t(ndraws,std::vector<tree>(m));
	for(size_t i=0;i<ndraws;i++)
		for(size_t j=0;j<m;j++)
			treef >> t[i][j];

	//Load the calibration parameter dimensions indices
	cout << "Loading calibration parameter indices" << endl;
  	std::stringstream cfss;
	std::string cfs;
	unsigned int ctemp, ncaldims;
	std::vector<unsigned int> caldims;
	cfss << "caldims.txt";
  	cfs=cfss.str();
  	std::ifstream cf(cfs.c_str()); //file to read caldims from for this slave
	while(cf >> ctemp) {
		caldims.push_back(ctemp-1);
		if(ctemp<0 || ctemp>p)
		{
			cout << "error: calibration dimensions is outside range of number of predictors p" << endl;
			return 1;
		}
	}
	ncaldims=caldims.size();

	// Load up the model/field indicator variable Im
	std::stringstream imfss;
	std::string imfs;
	unsigned int imdim;
	imfss << "imdim.txt";
	imfs=imfss.str();
#ifndef SILENT
	cout << ": loading model/field indicator variable Im from " << imfs << endl;
#endif
	std::ifstream imf(imfs.c_str()); //file to read Im from
	imf >> imdim;
	imdim--;
	if(imdim<0 || imdim>p)
	{
		cout << "error: Imdim dimension is outside the range of number of predictors p" << endl;
		return 1;
	}
#ifndef SILENT
	cout << ": loaded Imdim dimension Im=" << imdim << endl;
#endif

	//Load the thetas from the saved posterior
	cout << "Loading saved thetas from posterior" << endl;
	std::ifstream bcpf("bart-caldims.txt");
	double thetatemp;
	std::vector<double> thetas;
	while(bcpf >> thetatemp)
		thetas.push_back(thetatemp);

	size_t nthetas = thetas.size()/ncaldims;
	if(nthetas != ndraws) {
		cout << "error, wrong number of posterior samples saved in bart-caldims.txt\n";
		return 1;
	}

#ifndef SILENT
	cout << "Loaded theta posterior:" << endl;
	for(size_t i=0;i<nthetas;i++)
		cout << thetas[i] << endl;
#endif

	//Load the prediction points for each of the nodes
	cout << "Loading prediction points" << endl;
	if(argc<2) {cout << "error, expecting prediction root filename as first argument\n"; return 1;}
	double xtemp; //used to read in xpred without knowing how many there are
	char ptype='p'; //type of prediction, 'p' means mean of posterior predictive, 'd' means a single draw from
					//the posterior predictive
	if(argc>2) ptype=(char) *argv[2];
	cout << "Prediction Format Type = " << ptype << endl;
	dinfo dip;  //data information for prediction
	dip.n=0;
	dip.nf=0;
	std::vector<double> xp;  //prediction observations, stored like x
	std::stringstream xfss;
	std::string xfs;
	if(!ptype=='c')
		xfss << argv[1] << ".txt";
	else
		xfss << argv[1]  << ".txt";

	xfs=xfss.str();
	std::ifstream xpf(xfs.c_str()); //file to read x to predict at from for this node

	while(xpf >> xtemp)
		xp.push_back(xtemp);
	size_t np = xp.size()/p;
	if(xp.size() != np*p) {
		cout << "error, wrong number of elements in prediction data set\n";
		return 1;
	}
	if(np)
		dip.nf=np; dip.p=p; dip.xf=&xp[0]; dip.yf=0; //there are no y's!

#ifndef SILENT
	cout << "\nx for prediction read in:\n";
	cout "nf: " << dip.nf << endl;
#endif
	if(dip.nf) {
//#ifndef SILENT
		cout << " first row: " << dip.xf[0] << " ... " << dip.xf[p-1] << endl;
		cout << " last row: " << dip.xf[(dip.nf-1)*p] << " ... " << dip.xf[dip.nf*p-1] << endl;
//#endif
	}

	//Do this nodes predictions
	double* ppredmean=0; //posterior mean for prediction
	double* fpredmean=0; //posterior mean for field
	double* dpredmean=0; //posterior mean for the discrepancy
	double* fpredtemp=0; //temporary fit vector to compute prediction
	double* postp=0; //will hold the posterior mean over all the draws saved from the mcmc
	double* postp2=0; //will hold the posterior mean of E[f^2] over all the draws saved from the mcmc
	double* postf=0; //will hold the posterior mean for field over all the draws saved from the mcmc
	double* postf2=0; //will hold the posterior mean of E[f^2] over all the draws saved from the mcmc
	double* postd=0; //will hold the posterior mean for the discrepancy over all the draws saved from the mcmc
	double* postd2=0; //will hold the posterior mean of E[d^2] over all the draws saved from the mcmc
	ppredmean = new double[dip.nf];
	fpredmean = new double[dip.nf];
	dpredmean = new double[dip.nf];
  	fpredtemp = new double[dip.nf];
	postp = new double[dip.nf];
	postp2 = new double[dip.nf];
	postf = new double[dip.nf];
	postf2 = new double[dip.nf];
	postd = new double[dip.nf];
	postd2 = new double[dip.nf];
   	for(size_t i=0;i<dip.nf;i++) {
		ppredmean[i]=0.0;
		postp[i]=0.0;
		postp2[i]=0.0;
		fpredmean[i]=0.0;
		dpredmean[i]=0.0;
		postf[i]=0.0;
		postf2[i]=0.0;
		postd[i]=0.0;
		postd2[i]=0.0;
	}

	if(ptype=='p')
	{
		cout << "Calculating field prediction and standard deviation" << endl;
		size_t tdx=0;

		for(size_t i=0;i<ndraws;i++) {

			for(size_t k=0;k<ncaldims;k++) {
				for(size_t j=0;j<dip.nf;j++)
					xp[dip.p*j+caldims[k]]=thetas[tdx];
				tdx++;
			}

			//calculate prediction for field
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=1.0;
			makefieldpred(dip,xi,t[i],fpredmean);

			//calculate prediction for model
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=0.0;
			makefieldpred(dip,xi,t[i],ppredmean);

			//calculate prediction for discrepancy=field-model
			for(size_t j=0;j<dip.nf;j++)
				dpredmean[j]=fpredmean[j]-ppredmean[j];

			for(size_t k=0;k<dip.nf;k++) {
				if(i>0)
				{
					postp[k] *= (int)i;
					postp[k] += ppredmean[k];
					postp[k] /= (int)(i+1);
					postp2[k] *= (int)i;
					postp2[k] += ppredmean[k]*ppredmean[k];
					postp2[k] /= (int)(i+1);

					postf[k] *= (int)i;
					postf[k] += fpredmean[k];
					postf[k] /= (int)(i+1);
					postf2[k] *= (int)i;
					postf2[k] += fpredmean[k]*fpredmean[k];
					postf2[k] /= (int)(i+1);

					postd[k] *= (int)i;
					postd[k] += dpredmean[k];
					postd[k] /= (int)(i+1);
					postd2[k] *= (int)i;
					postd2[k] += dpredmean[k]*dpredmean[k];
					postd2[k] /= (int)(i+1);

				}
				else
				{
					postp[k] = ppredmean[k];
					postp2[k] = ppredmean[k]*ppredmean[k];
					postf[k] = fpredmean[k];
					postf2[k] = fpredmean[k]*fpredmean[k];
					postd[k] = dpredmean[k];
					postd2[k] = dpredmean[k]*dpredmean[k];
				}
				ppredmean[k]=0.0;
				fpredmean[k]=0.0;
				dpredmean[k]=0.0;
			}
		}
	
		//Write out this nodes model predictions
		std::stringstream yfps;
		std::string yfs;
		yfps << "yp.txt";
		yfs=yfps.str();
	#ifndef SILENT
		cout << "saving yp's to " << yfs << endl;
	#endif
		std::ofstream ypf(yfs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			ypf << postp[i] << endl;

		//Write out this nodes field predictions
		std::stringstream ffps;
		std::string ffs;
		ffps << "yfp.txt";
		ffs=ffps.str();
	#ifndef SILENT
		cout << "saving yf's to " << ffs << endl;
	#endif
		std::ofstream yff(ffs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			yff << postf[i] << endl;

		//Write out this nodes discrepancy predictions
		std::stringstream dfps;
		std::string dfs;
		dfps << "yd.txt";
		dfs=dfps.str();
	#ifndef SILENT
		cout << "saving yd's to " << dfs << endl;
	#endif
		std::ofstream ypd(dfs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			ypd << postd[i] << endl;

		//Write out this nodes model std deviations
		std::stringstream yfsd;
		std::string ysd;
		yfsd << "ysd.txt";
		ysd=yfsd.str();
	#ifndef SILENT
		cout <<"saving ysd's to " << ysd << endl;
	#endif
		std::ofstream ysdf(ysd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			ysdf << sqrt(postp2[i]-postp[i]*postp[i]) << endl;

		//Write out this nodes field std deviations
		std::stringstream ffsd;
		std::string fsd;
		ffsd << "yfsd.txt";
		fsd=ffsd.str();
	#ifndef SILENT
		cout <<"saving yfsd's to " << fsd << endl;
	#endif
		std::ofstream yfsdf(fsd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			yfsdf << sqrt(postf2[i]-postf[i]*postf[i]) << endl;

		//Write out this nodes discrepancy std deviations
		std::stringstream ydsd;
		std::string ydd;
		ydsd << "ydsd.txt";
		ydd=ydsd.str();
	#ifndef SILENT
		cout <<"saving ydsd's to " << ydd << endl;
	#endif
		std::ofstream ydsdf(ydd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			ydsdf << sqrt(postd2[i]-postd[i]*postd[i]) << endl;
	}
	else if(ptype=='m')  //Just predict the simulator outputs
	{
		cout << "Calculating simulator prediction and standard deviation" << endl;

		for(size_t i=0;i<ndraws;i++) {

			//calculate prediction for field
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=1.0;
			makefieldpred(dip,xi,t[i],fpredmean);

			//calculate prediction for model
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=0.0;
			makefieldpred(dip,xi,t[i],ppredmean);

			cout << " first row: " << dip.xf[0] << " " << dip.xf[1] << "..." << dip.xf[p-1] << endl;
			cout << " last row: " << dip.xf[(dip.nf-1)*p] << " " << dip.xf[dip.nf*p-2] << " ... " << dip.xf[dip.nf*p-1] << endl;

			//calculate prediction for discrepancy=field-model
			for(size_t j=0;j<dip.nf;j++)
				dpredmean[j]=fpredmean[j]-ppredmean[j];

			for(size_t k=0;k<dip.nf;k++) {
				if(i>0)
				{
					postp[k] *= (int)i;
					postp[k] += ppredmean[k];
					postp[k] /= (int)(i+1);
					postp2[k] *= (int)i;
					postp2[k] += ppredmean[k]*ppredmean[k];
					postp2[k] /= (int)(i+1);

					postf[k] *= (int)i;
					postf[k] += fpredmean[k];
					postf[k] /= (int)(i+1);
					postf2[k] *= (int)i;
					postf2[k] += fpredmean[k]*fpredmean[k];
					postf2[k] /= (int)(i+1);

					postd[k] *= (int)i;
					postd[k] += dpredmean[k];
					postd[k] /= (int)(i+1);
					postd2[k] *= (int)i;
					postd2[k] += dpredmean[k]*dpredmean[k];
					postd2[k] /= (int)(i+1);

				}
				else
				{
					postp[k] = ppredmean[k];
					postp2[k] = ppredmean[k]*ppredmean[k];
					postf[k] = fpredmean[k];
					postf2[k] = fpredmean[k]*fpredmean[k];
					postd[k] = dpredmean[k];
					postd2[k] = dpredmean[k]*dpredmean[k];
				}
				ppredmean[k]=0.0;
				fpredmean[k]=0.0;
				dpredmean[k]=0.0;
			}
		}
	
		//Write out this nodes model predictions
		std::stringstream yfps;
		std::string yfs;
		yfps << "yp.txt";
		yfs=yfps.str();
	#ifndef SILENT
		cout << "saving yp's to " << yfs << endl;
	#endif
		std::ofstream ypf(yfs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			ypf << postp[i] << endl;

		//Write out this nodes field predictions
		std::stringstream ffps;
		std::string ffs;
		ffps << "yfp.txt";
		ffs=ffps.str();
	#ifndef SILENT
		cout << "saving yf's to " << ffs << endl;
	#endif
		std::ofstream yff(ffs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			yff << postf[i] << endl;

		//Write out this nodes discrepancy predictions
		std::stringstream dfps;
		std::string dfs;
		dfps << "yd.txt";
		dfs=dfps.str();
	#ifndef SILENT
		cout << "saving yd's to " << dfs << endl;
	#endif
		std::ofstream ypd(dfs.c_str()); //file to save the predictions to for this node
		for(size_t i=0;i<dip.nf;i++)
			ypd << postd[i] << endl;

		//Write out this nodes model std deviations
		std::stringstream yfsd;
		std::string ysd;
		yfsd << "ysd.txt";
		ysd=yfsd.str();
	#ifndef SILENT
		cout <<"saving ysd's to " << ysd << endl;
	#endif
		std::ofstream ysdf(ysd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			ysdf << sqrt(postp2[i]-postp[i]*postp[i]) << endl;

		//Write out this nodes field std deviations
		std::stringstream ffsd;
		std::string fsd;
		ffsd << "yfsd.txt";
		fsd=ffsd.str();
	#ifndef SILENT
		cout <<"saving yfsd's to " << fsd << endl;
	#endif
		std::ofstream yfsdf(fsd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			yfsdf << sqrt(postf2[i]-postf[i]*postf[i]) << endl;

		//Write out this nodes discrepancy std deviations
		std::stringstream ydsd;
		std::string ydd;
		ydsd << "ydsd.txt";
		ydd=ydsd.str();
	#ifndef SILENT
		cout <<"saving ydsd's to " << ydd << endl;
	#endif
		std::ofstream ydsdf(ydd.c_str()); //file to save the std deviations to for this node
		for(size_t i=0;i<dip.nf;i++)
			ydsdf << sqrt(postd2[i]-postd[i]*postd[i]) << endl;
	}
	else if(ptype=='d') //write out the actual draws from the posterior
	{
		cout << "Writing draws from posterior" << endl;

		// file for model predictions
		std::stringstream yfps;
		std::string yfs;
		yfps << "yp.txt";
		yfs=yfps.str();
	#ifndef SILENT
		cout << "saving yp's to " << yfs << endl;
	#endif
		std::ofstream ypf(yfs.c_str()); //file to save the predictions to for this node

		// file for field predictions
		std::stringstream ffps;
		std::string ffs;
		ffps << "yfp.txt";
		ffs=ffps.str();
	#ifndef SILENT
		cout << "saving yf's to " << ffs << endl;
	#endif
		std::ofstream yff(ffs.c_str()); //file to save the predictions to for this node

		// file for discrepancy predictions
		std::stringstream dfps;
		std::string dfs;
		dfps << "yd.txt";
		dfs=dfps.str();
	#ifndef SILENT
		cout << "saving yd's to " << dfs << endl;
	#endif
		std::ofstream ypd(dfs.c_str()); //file to save the predictions to for this node

		size_t tdx=0;

		for(size_t i=0;i<ndraws;i++) {

			for(size_t k=0;k<ncaldims;k++) {
				for(size_t j=0;j<dip.nf;j++)
					xp[dip.p*j+caldims[k]]=thetas[tdx];
				tdx++;
			}
			//calculate prediction for field
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=1.0;
			makefieldpred(dip,xi,t[i],fpredmean);

			//calculate prediction for model
			for(size_t j=0;j<dip.nf;j++)
				xp[dip.p*j+imdim]=0.0;
			makefieldpred(dip,xi,t[i],ppredmean);


			//calculate prediction for discrepancy=field-model
			for(size_t j=0;j<dip.nf;j++)
				dpredmean[j]=fpredmean[j]-ppredmean[j];

			//write draws to files

			//Write out this nodes model predictions
			for(size_t i=0;i<dip.nf;i++)
				ypf << ppredmean[i] << " ";
			ypf << endl;

			//Write out this nodes field predictions
			for(size_t i=0;i<dip.nf;i++)
				yff << fpredmean[i] << " ";
			yff << endl;

			//Write out this nodes discrepancy predictions
			for(size_t i=0;i<dip.nf;i++)
				ypd << dpredmean[i] << " ";
			ypd << endl;
		}
	}

	//cleanup
	delete[] ppredmean;
	delete[] fpredmean;
	delete[] dpredmean;
	delete[] fpredtemp;
	delete[] postp;
	delete[] postp2;
	delete[] postf;
	delete[] postf2;
	delete[] postd;
	delete[] postd2;

	return 0;
} //end of main
